@@ -24,7 +24,6 @@ import (
 	"fmt"
 	"io"
 	"io/fs"
-	"io/ioutil"
 	"net/url"
 	"os"
 	"path"
@@ -43,7 +42,13 @@ import (
 
 const (
 	// DefaultRemoteRoot is the default remote TUF root location.
-	DefaultRemoteRoot = "https://sigstore-tuf-root.storage.googleapis.com"
+	DefaultRemoteRoot = "https://tuf-repo-cdn.sigstore.dev"
+	// defaultRemoteGCSBucket is the name of the GCS bucket that holds sigstore's public good production TUF root
+	defaultRemoteGCSBucket = "sigstore-tuf-root"
+	// defaultRemoteRootNoCDN is the URL of the GCS HTTP endpoint for the DefaultRootGCSBucket content
+	defaultRemoteRootNoCDN = "https://sigstore-tuf-root.storage.googleapis.com"
+	// defaultRemoteRootNoCDNAlt is an alternate URL to the GCS HTTP endpoint for the DefaultRootGCSBucket content
+	defaultRemoteRootNoCDNAlt = "https://storage.googleapis.com/sigstore-tuf-root"
 
 	// TufRootEnv is the name of the environment variable that locates an alternate local TUF root location.
 	TufRootEnv = "TUF_ROOT"
@@ -58,6 +63,9 @@ var (
 	singletonTUF     *TUF
 	singletonTUFOnce = new(sync.Once)
 	singletonTUFErr  error
+
+	// initMu locks concurrent calls to initializeTUF
+	initMu sync.Mutex
 )
 
 // getRemoteRoot is a var for testing.
@@ -73,6 +81,18 @@ type TUF struct {
 	mirror   string // location of mirror
 }
 
+// Mirror returns the mirror configured; note if the object was configured with a legacy reference
+// to the GCS HTTP endpoint for sigstore's public good trust root, this will return DefaultRemoteRoot
+// which is a CDN fronting that DefaultRemoteGCSBucket
+func (t *TUF) Mirror() string {
+	switch t.mirror {
+	case defaultRemoteGCSBucket, defaultRemoteRootNoCDN, defaultRemoteRootNoCDNAlt:
+		return DefaultRemoteRoot
+	default:
+		return t.mirror
+	}
+}
+
 // JSON output representing the configured root status
 type RootStatus struct {
 	Local    string                    `json:"local"`
@@ -96,6 +116,7 @@ type TargetFile struct {
 type customMetadata struct {
 	Usage  UsageKind  `json:"usage"`
 	Status StatusKind `json:"status"`
+	URI    string     `json:"uri"`
 }
 
 type sigstoreCustomMetadata struct {
@@ -173,7 +194,7 @@ func (t *TUF) getRootStatus() (*RootStatus, error) {
 	}
 	status := &RootStatus{
 		Local:    local,
-		Remote:   t.mirror,
+		Remote:   t.Mirror(),
 		Metadata: make(map[string]MetadataStatus),
 		Targets:  []string{},
 	}
@@ -240,6 +261,11 @@ func GetRootStatus(ctx context.Context) (*RootStatus, error) {
 // * forceUpdate: indicates checking the remote for an update, even when the local
 // timestamp.json is up to date.
 func initializeTUF(mirror string, root []byte, embedded fs.FS, forceUpdate bool) (*TUF, error) {
+	initMu.Lock()
+	defer initMu.Unlock()
+
+	// TODO: If a temporary error occurs for a long-running process, this singleton will
+	// never retry
 	singletonTUFOnce.Do(func() {
 		t := &TUF{
 			mirror:   mirror,
@@ -252,7 +278,7 @@ func initializeTUF(mirror string, root []byte, embedded fs.FS, forceUpdate bool)
 			return
 		}
 
-		t.remote, singletonTUFErr = remoteFromMirror(t.mirror)
+		t.remote, singletonTUFErr = remoteFromMirror(t.Mirror())
 		if singletonTUFErr != nil {
 			return
 		}
@@ -275,29 +301,35 @@ func initializeTUF(mirror string, root []byte, embedded fs.FS, forceUpdate bool)
 			}
 		}
 
-		if err := t.client.InitLocal(root); err != nil {
+		if err := t.client.Init(root); err != nil {
 			singletonTUFErr = fmt.Errorf("unable to initialize client, local cache may be corrupt: %w", err)
 			return
 		}
 
-		// We may already have an up-to-date local store! Check to see if it needs to be updated.
-		trustedTimestamp, ok := trustedMeta["timestamp.json"]
-		if ok && !isExpiredTimestamp(trustedTimestamp) && !forceUpdate {
-			// We're golden so stash the TUF object for later use
-			singletonTUF = t
-			return
-		}
-
-		// Update if local is not populated or out of date.
-		if err := t.updateMetadataAndDownloadTargets(); err != nil {
-			singletonTUFErr = fmt.Errorf("updating local metadata and targets: %w", err)
-			return
-		}
-
-		// We're golden so stash the TUF object for later use
 		singletonTUF = t
 	})
-	return singletonTUF, singletonTUFErr
+	if singletonTUFErr != nil {
+		return nil, singletonTUFErr
+	}
+
+	trustedMeta, err := singletonTUF.local.GetMeta()
+	if err != nil {
+		return nil, fmt.Errorf("getting trusted meta: %w", err)
+	}
+
+	// We may already have an up-to-date local store! Check to see if it needs to be updated.
+	trustedTimestamp, ok := trustedMeta["timestamp.json"]
+	if ok && !isExpiredTimestamp(trustedTimestamp) && !forceUpdate {
+		// We're golden so stash the TUF object for later use
+		return singletonTUF, nil
+	}
+
+	// Update if local is not populated or out of date.
+	if err := singletonTUF.updateMetadataAndDownloadTargets(); err != nil {
+		return nil, fmt.Errorf("updating local metadata and targets: %w", err)
+	}
+
+	return singletonTUF, nil
 }
 
 // TODO: Remove ctx arg.
@@ -316,15 +348,16 @@ func NewFromEnv(_ context.Context) (*TUF, error) {
 	return initializeTUF(mirror, nil, getEmbedded(), false)
 }
 
-func Initialize(ctx context.Context, mirror string, root []byte) error {
+func Initialize(_ context.Context, mirror string, root []byte) error {
 	// Initialize the client. Force an update with remote.
-	if _, err := initializeTUF(mirror, root, getEmbedded(), true); err != nil {
+	tuf, err := initializeTUF(mirror, root, getEmbedded(), true)
+	if err != nil {
 		return err
 	}
 
 	// Store the remote for later if we are caching.
 	if !noCache() {
-		remoteInfo := &remoteCache{Mirror: mirror}
+		remoteInfo := &remoteCache{Mirror: tuf.Mirror()}
 		b, err := json.Marshal(remoteInfo)
 		if err != nil {
 			return err
@@ -337,15 +370,17 @@ func Initialize(ctx context.Context, mirror string, root []byte) error {
 }
 
 // Checks if the testTarget matches the valid target file metadata.
-func isValidTarget(testTarget []byte, validMeta data.TargetFileMeta) bool {
-	localMeta, err := util.GenerateTargetFileMeta(bytes.NewReader(testTarget))
+func isValidTarget(testTarget []byte, validMeta data.TargetFileMeta) (bool, error) {
+	localMeta, err := util.GenerateTargetFileMeta(
+		bytes.NewReader(testTarget),
+		"sha256", "sha512")
 	if err != nil {
-		return false
+		return false, err
 	}
 	if err := util.TargetFileMetaEqual(localMeta, validMeta); err != nil {
-		return false
+		return false, err
 	}
-	return true
+	return true, nil
 }
 
 func (t *TUF) GetTarget(name string) ([]byte, error) {
@@ -361,8 +396,8 @@ func (t *TUF) GetTarget(name string) ([]byte, error) {
 		return nil, err
 	}
 
-	if !isValidTarget(targetBytes, validMeta) {
-		return nil, fmt.Errorf("cache contains invalid target; local cache may be corrupt")
+	if valid, err := isValidTarget(targetBytes, validMeta); !valid {
+		return nil, fmt.Errorf("cache contains invalid target; local cache may be corrupt: %w", err)
 	}
 
 	return targetBytes, nil
@@ -417,37 +452,7 @@ func (t *TUF) GetTargetsByMeta(usage UsageKind, fallbacks []string) ([]TargetFil
 func (t *TUF) updateClient() (data.TargetFiles, error) {
 	targets, err := t.client.Update()
 	if err != nil {
-		// Get some extra information for debugging. What was the state of the top-level
-		// metadata on the remote?
-		status := struct {
-			Mirror   string                    `json:"mirror"`
-			Metadata map[string]MetadataStatus `json:"metadata"`
-		}{
-			Mirror:   t.mirror,
-			Metadata: make(map[string]MetadataStatus),
-		}
-		for _, md := range []string{"root.json", "targets.json", "snapshot.json", "timestamp.json"} {
-			r, _, err := t.remote.GetMeta(md)
-			if err != nil {
-				// May be missing, or failed download.
-				continue
-			}
-			defer r.Close()
-			b, err := ioutil.ReadAll(r)
-			if err != nil {
-				continue
-			}
-			mdStatus, err := getMetadataStatus(b)
-			if err != nil {
-				continue
-			}
-			status.Metadata[md] = *mdStatus
-		}
-		b, innerErr := json.MarshalIndent(status, "", "\t")
-		if innerErr != nil {
-			return nil, innerErr
-		}
-		return nil, fmt.Errorf("error updating to TUF remote mirror: %w\nremote status:%s", err, string(b))
+		return nil, fmt.Errorf("error updating to TUF remote mirror: %w", err)
 	}
 	// Success! Cache new metadata, if needed.
 	if noCache() {
@@ -507,7 +512,7 @@ func maybeDownloadRemoteTarget(name string, meta data.TargetFileMeta, t *TUF) er
 	// If we already have the target locally, don't bother downloading from remote storage.
 	if cachedTarget, err := t.targets.Get(name); err == nil {
 		// If the target we have stored matches the meta, use that.
-		if isValidTarget(cachedTarget, meta) {
+		if valid, _ := isValidTarget(cachedTarget, meta); valid {
 			return nil
 		}
 	}
@@ -528,7 +533,7 @@ func maybeDownloadRemoteTarget(name string, meta data.TargetFileMeta, t *TUF) er
 			b = bytes.ReplaceAll(b, []byte("\r\n"), []byte("\n"))
 		}
 
-		if isValidTarget(b, meta) {
+		if valid, _ := isValidTarget(b, meta); valid {
 			if _, err := io.Copy(&w, bytes.NewReader(b)); err != nil {
 				return fmt.Errorf("using embedded target: %w", err)
 			}
@@ -677,10 +682,12 @@ func (d *diskCache) Set(p string, b []byte) error {
 	if err := d.memory.Set(p, b); err != nil {
 		return err
 	}
-	if err := os.MkdirAll(d.base, 0o700); err != nil {
+
+	fp := filepath.FromSlash(filepath.Join(d.base, p))
+	if err := os.MkdirAll(filepath.Dir(fp), 0o700); err != nil {
 		return fmt.Errorf("creating targets dir: %w", err)
 	}
-	fp := filepath.FromSlash(filepath.Join(d.base, p))
+
 	return os.WriteFile(fp, b, 0o600)
 }
 
@@ -694,8 +701,13 @@ func noCache() bool {
 
 func remoteFromMirror(mirror string) (client.RemoteStore, error) {
 	// This is for compatibility with specifying a GCS bucket remote.
-	if _, parseErr := url.ParseRequestURI(mirror); parseErr != nil {
-		mirror = fmt.Sprintf("https://%s.storage.googleapis.com", mirror)
+	u, parseErr := url.ParseRequestURI(mirror)
+	if parseErr != nil {
+		return client.HTTPRemoteStore(fmt.Sprintf("https://%s.storage.googleapis.com", mirror), nil, nil)
 	}
-	return client.HTTPRemoteStore(mirror, nil, nil)
+	if u.Scheme != "file" {
+		return client.HTTPRemoteStore(mirror, nil, nil)
+	}
+	// Use local filesystem for remote.
+	return client.NewFileRemoteStore(os.DirFS(u.Path), "")
 }
